import lcapy.grammar as grammar
import lcapy.mnacpts as mnacpts
import pytest
from lcapy.parser import Parser
import sys

sys.path.append('..')

parser = Parser(mnacpts, grammar)
parse = parser.parse


def test_Exception1():
    """Test missing arg"""
    with pytest.raises(ValueError):
        parse('V1 2')


def test_Exception2():
    """Test too many args"""
    with pytest.raises(ValueError):
        parse('V1 2 3 4 5 6')


def test_Exception3():
    """Test too many args"""
    with pytest.raises(ValueError):
        parse('V1 2 3 dc 4 5 6')


def test_Exception4():
    """Test unknown component"""

    with pytest.raises(ValueError):
        parse('B1 1 2')


def test_Exception5():
    """Test arg reassignment"""
    with pytest.raises(ValueError):
        parse('E1 1 2 opamp 3 4 A Ad=10')


def test_Exception6():
    """Test unknown arg"""
    with pytest.raises(ValueError):
        parse('V1 1 2 foo=3')


def test_V1():
    """Test voltage source"""

    assert type(parse('V1 1 2')) == mnacpts.V, 'Class not V'
    assert str(parse('V1 1 2')) == 'V1 1 2', 'V netmake'


def test_Vdc1():
    """Test dc voltage source"""

    assert type(parse('V1 1 2 dc')) == mnacpts.classes['Vdc'], 'Class not Vdc'
    assert str(parse('V1 1 2 dc')) == 'V1 1 2 dc', 'V dc netmake'


def test_Vac1():
    """Test ac voltage source"""

    assert type(parse('V1 1 2 ac')) == mnacpts.classes['Vac'], 'Class not Vac'


def test_Vsin1():
    """Test sin voltage source"""

    assert type(parse('V1 1 2 sin(1, 2, 3)')) == mnacpts.classes['Vsin'], 'Class not Vsin'


def test_Vexpr():
    """Test voltage source with expr"""

    assert type(parse('V1 1 2 {a * 5}')) == mnacpts.V, 'Class not V'
    assert str(parse('V1 1 2 {a * 5}')) == 'V1 1 2 {a * 5}', 'V expr netmake'


def test_Vexpr2():
    """Test voltage source with expr as name"""

    assert str(parse('V1 1 2 V1')) == 'V1 1 2', 'V expr2 netmake'


def test_Vac2():
    """Test ac voltage source"""

    assert type(parse('V1 1 2 ac {1j} {0} 3')) == mnacpts.classes['Vac'], 'Class not Vac'


def test_Vquotes():
    """Test voltage source with arg in quotes"""

    assert type(parse('V1 1 2 "a * 5"')) == mnacpts.V, 'Class not V'

# def test_opamp():
#     '''Test opamp'''
#
#     assert type(parse('E 1 2 opamp 3 4')) == mnacpts.classes['Eopamp'], 'Class not Eopamp'



"""This module handles the nodes for a circuit.

Copyright 2023 Michael Hayes, UCECE
"""

from .attrdict import AttrDict
from .schemmisc import Pos
from .node import Node


class Nodes(AttrDict):
    """An instance of this stores all the defined Nodes for a circuit as a
    dictionary.  Elements in the direction can also be accessed as
    attributes.  Each Node has a list of associated components.

    """

    def add(self, node_name, cpt, cct):
        """Add a new node with associated component and circuit."""

        if node_name in self:
            node = self[node_name]
        else:
            node = Node(cct, node_name)
            self[node_name] = node
        node.append(cpt)
        return node

    def remove(self, node_name, cpt):
        """Detach node from component and delete the node
        if there are no more uses."""

        if node_name not in self:
            raise ValueError('Cannot remove unknown node', node_name)
        node = self[node_name]
        node.remove(cpt)

    def rename(self, old_node_name, new_node_name, cpts=None):
        """Rename old node name to new node name if the old node is connected
        to a component in the list cpts or if cpts is None.  Returns
        the new node."""

        if old_node_name not in self:
            raise ValueError('Cannot rename unknown node', old_node_name)
        node = self[old_node_name]
        return node.rename(new_node_name, cpts)

    def _delete(self, node_name):
        """Delete node from dict of nodes.  Note, it is best to use remove
        since this will not delete the Node until all uses are
        removed."""

        node = self[node_name]
        if len(node.connected) != 0:
            raise ValueError('Cannot remove node ' + str(node_name) +
                             '; it is needed for ' +
                             ', '.join([str(name) for name in node.connected]))

        self.pop(node_name)

    def debug(self):

        s = ''
        for node in self.values():
            s += node.debug()
        return s

    def by_position(self, position):

        x, y = position

        for node in self.values():
            if abs(node.x - x) < 1e-5 and abs(node.y - y) < 1e-5:
                return node
        return None


def parse_nodes(nodesstr):

    from .utils import split_parens

    node_positions = {}

    # Ignore {}
    nodesstr = nodesstr[1:-1]
    entries = split_parens(nodesstr, ',')
    for entry in entries:
        if entry == '':
            continue
        parts = entry.split('@')
        node_name = parts[0].strip()
        # Ignore ()
        values = parts[1][1:-1]
        parts = values.split(',')
        x = float(parts[0])
        y = float(parts[1])
        pos = Pos(x, y)

        node_positions[node_name] = pos

    return node_positions



from .quadraticroot import QuadraticRoot


class Root:

    def __init__(self, expr, n, damping=None):

        self.n = n

        d = None
        if damping is not None:
            # Perhaps parse as quadratic root, cubic root, etc.?
            d = QuadraticRoot.from_expr(expr, damping)

        if d is None:
            self.expr = expr
            self.conj = expr.conjugate()
        else:
            self.expr = d.expr
            self.conj = d.conjugate()

        self.damping = damping
        self._d = d

    def conjugate(self):
        return self.conj

    def is_conjugate_pair(self, root):

        return self.expr == root.conj


def pair_conjugates(roots_dict, damping=None):
    """Return dictionary of conjugate root pairs and a dictionary of the
    remaining single roots."""

    root_single_dict = roots_dict.copy()
    root_pair_dict = {}

    root_list = list(roots_dict)

    P = {}
    for root in root_list:
        P[root] = Root(root, 1, damping=damping)

    for i, root in enumerate(root_list):

        p = P[root]

        for root_c in root_list[i + 1:]:

            pc = P[root_c]
            if p.is_conjugate_pair(pc):

                root_single_dict.pop(root, None)
                root_single_dict.pop(root_c, None)

                o1 = roots_dict[root]
                o2 = roots_dict[root_c]
                if o1 == o2:
                    root_pair_dict[root, root_c] = o1
                elif o1 > o2:
                    root_pair_dict[root, root_c] = o2
                    root_single_dict[root] = o1 - o2
                else:
                    root_pair_dict[root, root_c] = o1
                    root_single_dict[root_c] = o2 - o1

    return root_pair_dict, root_single_dict



"""This module provides support for Hilbert transforms.  It calculates
the Hilbert transform using:


Copyright 2024 Michael Hayes, UCECE

"""

from sympy import pi, oo, exp, sin, cos, log, sign, Integral, I, DiracDelta
from .sym import symsimplify, j, tausym
from .transformer import BilateralForwardTransformer
from .utils import factor_const
from .extrafunctions import rect

__all__ = ('HT', )


# If g(t) has a Hilbert transform H g(t) then
# g(t - t0) has a Hilbert transform H g(t - t0) and
# g(a t)has a Hilbert transform sign(a) H g(a t).
#
# H(g(t) * h(t)) = H g(t) * h(t) = g(t) * H h(t).
#
# H dg(t)/dt = d H g(t) dt




class HilbertTransformer(BilateralForwardTransformer):

    name = 'Hilbert transform'

    def key(self, expr, t, f, **kwargs):
        return expr, t, f

    def simplify_term(self, expr, var):

        return symsimplify(expr)

    def term(self, expr, t, f=None):

        const, expr = factor_const(expr, t)

        if self.is_inverse:
            const = -const

        # Check for constant.
        if not expr.has(t):
            return 0

        if expr.is_Pow and expr.args[1] == -1:
            # Handle 1 / t
            arg = expr.args[0]
            if arg.is_polynomial(t) and arg.as_poly(t).is_linear:
                c1 = arg.expand().coeff(t, 1)

                return pi * DiracDelta(arg) / c1

        if len(expr.args) == 0:
            return const * self.noevaluate(expr, t)

        arg = expr.args[0]

        if not arg.is_polynomial(t) or not arg.as_poly(t).is_linear:
            return const * self.noevaluate(expr, t)

        if not expr.is_Function:
            return const * self.noevaluate(expr, t)

        c1 = arg.expand().coeff(t, 1)

        if expr.func == DiracDelta:
            return const * sign(c1) * 1 / (pi * arg)

        if expr.func == cos:
            return const * sign(c1) * sin(arg)

        elif expr.func == sin:
            return const * sign(c1) * -cos(arg)

        elif expr.func == rect:
            return const * sign(c1) * -log(abs((arg - 1 / 2) / (arg + 1 / 2))) / pi

        elif expr.func == exp and c1.has(I):
            return const * sign(c1 / I) * -I * expr

        # TODO, add sinc pulse, Cauchy pulse, 1 / t


        q =  Integral(expr / (pi * (t - tausym)), (tausym, -oo, oo))
        return const * q.principal_value()

    def noevaluate(self, expr, t, f=None):

        return Integral(expr / (pi * (t - tausym)), (tausym, -oo, oo))

    def check(self, expr, t, f=None):
        pass

    def rewrite(self, expr, var):
        return expr


hilbert_transformer = HilbertTransformer()


def HT(expr, t, **kwargs):
    """Compute Hilbert transform of expr."""

    return hilbert_transformer.transform(expr, t, **kwargs)


def hilbert_transform(expr, t, **kwargs):
    """Compute Hilbert transform of expr."""

    return hilbert_transformer.transform(expr, t, **kwargs)



class Label:

    def __init__(self, key, val):

        # Circuitikz does not handle \\ in labels.  As a hack
        # convert l to l2 and a to a2.   Note, l2 and a2 can only
        # support two lines.
        parts = val.split(r'\\', 1)
        if len(parts) == 2:
            if key[0] == 'l':
                key = key.replace('l', 'l2')
                val = ' and '.join(parts)
            elif key[0] == 'a':
                key = key.replace('a', 'a2')
                val = ' and '.join(parts)

        self.key = key
        self.val = val

    def __str__(self):

        return self.key + '=' + self.val



from .label import Label
from .latex import latex_format_label


class Labels:

    def __init__(self):

        self.current = None
        self.voltage = None
        self.flow = None
        self.inner = None
        self.annotation = None
        self.label = None

    def add(self, key, val):

        val = latex_format_label(val)

        if key[0] == 'a':
            self.annotation = Label(key, val)
        elif key[0] == 'f':
            self.flow = Label(key, val)
        elif key[0] == 'i':
            if key == 'ir':
                key = 'i' + '<'
            self.current = Label(key, val)
        elif key[0] == 'l':
            self.label = Label(key, val)
        elif key[0] == 't':
            self.inner = Label(key, val)
        elif key[0] == 'v':
            if key == 'vr':
                key = key + '>'
            self.voltage = Label(key, val)

    def args(self, flip=False):

        def fmt2(label, flip=False):
            key, val = label.key, label.val

            if '^' not in key and '_' not in key:
                key = key + ('^' if flip else '_')

            return key + '=' + val

        def fmt(label):
            key, val = label.key, label.val
            return key + '=' + val

        args = []
        if self.annotation is not None:
            args.append(fmt2(self.annotation, not flip))
        if self.voltage is not None:
            args.append(fmt2(self.voltage, not flip))
        if self.flow is not None:
            args.append(fmt(self.flow))
        if self.current is not None:
            args.append(fmt(self.current))
        if self.inner is not None:
            args.append(fmt(self.inner))
        if self.label is not None:
            args.append(fmt2(self.label, flip))
        return args

    def __str__(self):

        args = self.args()
        return ', '.join(args)



from matplotlib.pyplot import savefig
from lcapy import s

r1 = 500
i1 = 500

r2 = 5000
i2 = 5000

p1 = r1 + 1j * i1
p1c = r1 - 1j * i1

p2 = r2 + 1j * i2
p2c = r2 - 1j * i2

H = (s + 2000)**2 / (s + p1) / (s + p1c) / (s + p2) / (s + p2c)

H.plot()
savefig(__file__.replace('.py', '.png'), bbox_inches='tight')



from sympy import sqrt, S, I

Zero = S.Zero
One = S.One


def _is_a_minus_b(expr):

    if not expr.is_Add:
        return False

    sub = 0
    add = 0
    for term in expr.as_ordered_terms():
        if ((term.is_Mul and term.args[0] < 0) or
            (term.is_constant() and term.is_negative)):
            sub += 1
        else:
            add += 1

    return add >= 1 and sub >= 1


def _decompose_bar(expr):

    # Look for scale * sqrt(dexpr) where dexpr = aexpr - bexpr

    dexpr = None
    scale = One

    for factor in expr.as_ordered_factors():
        if (factor.is_Pow and factor.args[1] == One / 2 and
            _is_a_minus_b(factor.args[0])):
            if dexpr is not None:
                return None
            dexpr = factor.args[0]
        else:
            scale *= factor
    if dexpr is None:
        return None

    return scale, dexpr


def _decompose_foo(expr):

    # Look for offset + scale * sqrt(dexpr) where dexpr = aexpr - bexpr

    offset = Zero
    scale = One
    dexpr = None

    for term in expr.as_ordered_terms():
        p = _decompose_bar(term)
        if p is None:
            offset += term
        elif dexpr is not None:
            return None
        else:
            scale, dexpr = p

    return offset, scale, dexpr


def _decompose(expr):

    scale = One
    scale2 = One
    offset = One
    dexpr = None

    for factor in expr.as_ordered_factors():
        if factor.is_Add:
            p = _decompose_foo(factor)
            if p is None:
                scale *= factor
            elif dexpr is not None:
                return None
            else:
                offset, scale2, dexpr = p
        else:
            scale *= factor

    if dexpr is None:
        dexpr = Zero
    return scale * offset, scale * scale2, dexpr


class QuadraticRoot:

    def __init__(self, offset, scale, dexpr, damping=None):

        # If the expression cannot be decomposed, then
        # scale = 1 and dexpr = 0.

        self.offset = offset
        self.scale = scale
        self.dexpr = dexpr
        self.damping = damping

    @property
    def expr(self):

        return self.offset + self.scale * sqrt(self.dexpr)

    def conjugate(self):

        if self.scale.is_imaginary:
            return self.offset - self.scale * sqrt(self.dexpr)
        else:
            return self.expr.conjugate()

    @classmethod
    def from_expr(cls, expr, damping=None):

        offset, scale, dexpr = _decompose(expr)

        if damping is None:
            pass
        elif damping == 'over':
            if scale.is_imaginary:
                dexpr = -dexpr
                scale /= I
        elif damping == 'critical':
            # This puts constraints on variables in dexpr.  For example,
            # if have sqrt(a - b) then a = b.
            pass
        elif damping == 'under':
            if not scale.is_imaginary:
                dexpr = -dexpr
                scale *= I
        else:
            raise ValueError('Unknown damping %s' % damping)

        if dexpr == 0:
            return None

        if damping == 'critical':
            dexpr = 0

        return cls(offset, scale, dexpr, damping)

    def is_conjugate_pair(self, d):

        if self.dexpr == 0 or d.dexpr == 0:
            return self.expr == d.expr.conjugate()

        if not self.scale.is_imaginary:
            return False

        if not d.scale.is_imaginary:
            return False

        return (self.offset == d.offset and self.scale == -d.scale and
                self.dexpr == d.dexpr)



from lcapy import *
from lcapy.expr import TimeDomainExpression
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_hilbert(self):

        self.assertEqual(DiracDelta(t).HT(), 1 / (pi * t), "delta(t)")
        self.assertEqual((1 / t).HT(), pi * DiracDelta(t), "1 / t")
        self.assertEqual(cos(t).HT(), sin(t), "cos(t)")
        self.assertEqual(sin(t).HT(), -cos(t), "cos(t)")
        self.assertEqual(exp(j * t).HT(), -j * exp(j * t), "exp(j * t)")
        self.assertEqual(exp(-j * t).HT(), j * exp(-j * t), "exp(-j * t)")

    def test_inverse_hilbert(self):

        self.assertEqual((1 / t).IHT(), pi * DiracDelta(t), "-1 / t")



"""This module provides support for nodes.

Copyright 2020--2024 Michael Hayes, UCECE

"""

from .immittancemixin import ImmittanceMixin
from copy import copy


class DummyNode:

    def __init__(self, name):
        self.name = name


class Node(ImmittanceMixin):
    """Each node has an instance of this.   """

    def __init__(self, cct, name):

        self.cct = cct
        self._name = name
        self.pos = None
        parts = name.split('_')
        self.rootname = parts[0] if name[0] != '_' else name
        self.primary = len(parts) == 1
        # List of components connected to this node.
        self._connected = []
        self._count = 0

    def __repr__(self):
        return "Node('%s')" % self.name

    def __str__(self):

        if self.pos is not None:
            # Note, need to have float type otherwise 0 becomes empty string.
            x = str(round(self.x, 2)).rstrip('0').rstrip('.')
            y = str(round(self.y, 2)).rstrip('0').rstrip('.')

            return '%s@(%s, %s)' % (self.name, x, y)
        else:
            return self.name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        """Rename node name."""

        self.rename(name)

    @property
    def x(self):

        if self.pos is None:
            raise ValueError('Node position undefined')
        return self.pos.x

    @property
    def y(self):

        if self.pos is None:
            raise ValueError('Node position undefined')
        return self.pos.y

    @property
    def V(self):
        """Node voltage with respect to ground."""

        return self.cct.get_Vd(self.name, '0')

    @property
    def v(self):
        """Node time-domain voltage with respect to ground."""

        return self.cct.get_vd(self.name, '0')

    @property
    def admittance(self):
        """Driving-point admittance between node and ground."""

        return self.cct.admittance(self.name, '0')

    @property
    def impedance(self):
        """Driving-point impedance between node and ground."""

        return self.cct.impedance(self.name, '0')

    def oneport(self, node=0):
        """Create oneport object with respect to specified node
        (default ground)."""

        return self.cct.oneport(self.name, node)

    def thevenin(self, node=0):
        """Create Thevenin oneport object with respect to specified
        node (default ground)."""

        return self.cct.thevenin(self.name, node)

    def norton(self, node=0):
        """Create Norton oneport object with respect to specified node
        (default ground)."""

        return self.cct.norton(self.name, node)

    @property
    def connected(self):
        """Return list of components connected to the node."""

        return self._connected

    @property
    def connected_nodes(self):
        """Return list of nodes that are connected to the node
        by a component."""

        nodes = []
        for cpt in self._connected:
            for node in cpt.nodes:
                if node is not self and node not in nodes:
                    nodes.append(node)
        return nodes

    def is_connected(self, cpt):
        """Return True if cpt is connected to the node."""

        if isinstance(cpt, str):
            for cpt1 in self.connected:
                if cpt1.name == cpt:
                    return True
            return False

        return cpt in self.connected

    def is_wired_to(self, node):
        """Return True if the nodes are wired together."""

        node = str(node)
        for nodes in self.cct.equipotential_nodes.values():
            if self.name in nodes and node in nodes:
                return True
        return False

    def wired_to(self):
        """Return list of names of nodes that are wired to this node."""

        for nodes in self.cct.equipotential_nodes.values():
            if self.name in nodes:
                nodes.remove(self.name)
                return nodes
        return []

    @property
    def is_port(self):
        """Return True if node is a port."""

        for cpt in self.connected:
            if cpt.type == 'P':
                return True
        return False

    @property
    def port(self):
        """Alias for is_port."""

        return self.is_port

    @property
    def is_dangling(self):
        """Return True if node has a single electrical connection"""

        return self._count <= 1

    @property
    def is_ground(self):
        """Return True if node is a ground."""

        return self.name.startswith('0')

    def append(self, cpt):
        """Attach component to the node."""

        if cpt.type not in ('A', 'O'):
            self._count += 1

        self._connected.append(cpt)

    def clone(self, name):
        """Create a copy of the node but with a new name."""

        node = Node(self.cct, name)
        node.pos = copy(self.pos)
        return node

    def remove(self, cpt):
        """Detach component from node.  If no other components
        are attached, the node is deleted."""

        if len(self._connected) == 0:
            raise RuntimeError(
                'Removing node %s with no connections' % str(self))

        for c in self._connected:
            if c.name == cpt.name:
                self._connected.remove(c)
                break
        if cpt.type not in ('A', 'O'):
            self._count -= 1

        if self.count == 0:
            self.cct.nodes._delete(self.name)

    def _rename1(self, name):
        # Case 1.  The name is new and is applied to all components

        if name in self.cct.nodes:
            raise ValueError('Node %s is not new' % name)

        self.cct.nodes[name] = self.cct.nodes.pop(self._name)
        self._name = name
        return self

    def _rename2(self, name):
        # Case 2.  The name is not new and is applied to all components

        if name not in self.cct.nodes:
            raise ValueError('Node %s is new' % name)

        self._connected.extend(self.cct.nodes[name]._connected)
        self._count += self.cct.nodes[name]._count

        self.cct.nodes[name] = self.cct.nodes.pop(self._name)
        self._name = name
        return self

    def _rename3(self, name, cpts):
        # Case 3.  The name is new and is applied to some of the components

        if name in self.cct.nodes:
            raise ValueError('Node %s is not new' % name)

        new_node = self.clone(name)
        self.cct.nodes[name] = new_node

        for cpt in cpts:
            for m, node in enumerate(cpt.nodes):
                if node.name == self.name:
                    if cpt.type not in ('A', 'O'):
                        self._count -= 1
                    new_node.append(cpt)
                    cpt.nodes[m] = new_node
        return new_node

    def _rename4(self, name, cpts):
        # Case 4.  The name is not new and is applied to some of the components

        if name not in self.cct.nodes:
            raise ValueError('Node %s is new' % name)

        new_node = self.cct.nodes[name]
        self.cct.nodes[name] = new_node

        for cpt in cpts:
            for m, node in enumerate(cpt.nodes):
                if node.name == self.name:
                    if cpt.type not in ('A', 'O'):
                        self._count -= 1
                    new_node.append(cpt)
                    cpt.nodes[m] = new_node
        return new_node

    def rename(self, name, cpts=None):
        """Rename node name if connected to a component in the list cpts or if
        cpts is None.  Return new node."""

        # Case 1. The name is new and is applied to all components
        # Case 2. The name is not new and is applied to all components
        # Case 3. The name is new and is applied to some of the components
        # Case 4. The name is not new and is applied to some of the components
        #
        # all and some applies to the components that are connected
        # to this node

        if cpts is None:
            if name not in self.cct.nodes:
                # Case 1.
                return self._rename1(name)
            else:
                # Case 2.
                return self._rename2(name)

        # Convert names to cpts
        if not isinstance(cpts, (list, tuple)):
            cpts = [cpts]

        tcpts = []
        for cpt in cpts:
            if isinstance(cpt, str):
                tcpts.append(self.cct[cpt])
            else:
                tcpts.append(cpt)
        cpts = tcpts

        other = []
        for cpt in self.connected:
            if cpt not in cpts:
                other.append(cpt)

        if other == []:
            # Case 1
            if name not in self.cct.nodes:
                return self._rename1(name)

            # Case 2
            return self._rename2(name)

        # Case 3
        if name not in self.cct.nodes:
            return self._rename3(name, cpts)

        # Case 4
        return self._rename4(name, cpts)

    @property
    def count(self):
        """Number of components electrically connected to the node.  This
        includes wires but not annotations.

        """

        return self._count

    def debug(self):

        s = str(self) + ', count=%s' % self.count

        names = [cpt.name for cpt in self.connected]

        s += ', cpts=[%s]' % ', '.join(names) + '\n'

        return s



"""This module provides support for inverse Hilbert transforms.

Copyright 2024 Michael Hayes, UCECE
"""

from .hilbert import HilbertTransformer
from .sym import j, pi
import sympy as sym


class InverseHilbertTransformer(HilbertTransformer):

    name = 'inverse Hilbert transform'
    is_inverse = True

    def noevaluate(self, expr, f, t):

        return sym.Integral(expr * sym.exp(j * 2 * pi * f * t), (f, -sym.oo, sym.oo))


inverse_hilbert_transformer = InverseHilbertTransformer()


def IHT(expr, t, **kwargs):
    """Compute inverse Hilbert transform of expr.

    Note, if the original function has a constant then
    it cannot be recovered."""

    return inverse_hilbert_transformer.transform(expr, t, **kwargs)


def inverse_hilbert_transform(expr, t, **kwargs):
    """Compute inverse Hilbert transform of expr.

    Note, if the original function has a constant then
    it cannot be recovered."""

    return inverse_hilbert_transformer.transform(expr, t, **kwargs)



from .expr import Expr
from .latex import latex_format_label
from .engformatter import EngFormatter
import sympy as sym


class LabelMaker:

    def _format_expr(self, expr):

        return Expr(expr, cache=False).latex_math()

    def _format_value_units(self, value, units):

        return EngFormatter().latex_math(value, units)

    def _format_name(self, cpt_type, cpt_id):

        name = cpt_type
        subscript = cpt_id[1:] if cpt_id.startswith('_') else cpt_id

        if cpt_type == 'REL':
            name = r'\mathcal{R}'
        elif len(name) > 1:
            name = r'\mathrm{%s}' % name

        if subscript != '':
            if len(subscript) > 1:
                subscript = r'\mathrm{%s}' % subscript
            name = name + '_{%s}' % subscript
        return latex_format_label('$' + name + '$')

    def make(self, cpt, label_ports=False):

        # There are two possible labels for a component:
        # 1. Component name, e.g., R1
        # 2. Component value, expression, or symbol

        id_label = self._format_name(cpt.type, cpt.id)
        value_label = None

        if cpt.type == 'P' and not label_ports:
            id_label = None

        elif cpt.type in ('A', 'O', 'W') or id_label.find('#') != -1:
            id_label = None

        if cpt.type in ('A', 'S', 'SW', 'U'):
            value_label = ''

        unify = False
        if len(cpt.args):

            # TODO, extend for mechanical and acoustical components.
            units_map = {'V': 'V', 'I': 'A', 'R': '$\Omega$',
                         'C': 'F', 'L': 'H'}

            expr = cpt.args[0]
            if cpt.classname in ('Vstep', 'Istep'):
                expr = '(%s) * Heaviside(t)' % expr
                value_label = self._format_expr(expr)
            elif cpt.classname in ('Vs', 'Is'):
                value_label = self._format_expr(expr)
            elif cpt.classname == 'TF':
                expr = sym.sympify(expr)
                if expr.is_Pow and expr.args[1] == -1:
                    value_label = '%s:1' % (1 / expr)
                else:
                    value_label = '1:%s' % expr
            elif cpt.type in ('F', 'H') and len(cpt.args) > 1:
                # This is hard to give a reasonable label since the
                # control current is specified by a voltage source.
                # The user will have to override manually.
                expr = cpt.args[1]
                value_label = self._format_expr(expr)
            elif cpt.classname not in ('TP',):
                try:
                    # Handle things like 9/1000 that can occur
                    # when substituting cpt values.
                    value = float(sym.Rational(expr))
                    if cpt.type in units_map:
                        value_label = self._format_value_units(
                            value, units_map[cpt.type])
                    else:
                        value_label = self._format_expr(expr)

                except (ValueError, TypeError):
                    # This catches non numeric arg.
                    value_label = self._format_expr(expr)

            # Ensure labels are the same when the value is not specified.
            # This will prevent printing the name and value.
            unify = expr == cpt.type + cpt.id

            # Currently, we only annnotate the component with the
            # value, expression, or symbol.  If this is not specified,
            # it defaults to the component identifier.  Note, there
            # are some objects we do not want to label, such as wires
            # and ports.
        id_label = '' if id_label is None else latex_format_label(id_label)
        value_label = id_label if value_label is None \
            else latex_format_label(value_label)

        if unify:
            value_label = id_label

        return id_label, value_label
